BUFFER_END_STYLES = {
    'ROUND': 1,
    'FLAT': 2,
    'SQUARE': 3
}

BUFFER_JOIN_STYLES = {
    'ROUND': 1,
    'MITRE': 2,
    'BEVEL': 3
}

METADATA_TABLES = [
    'spatial_ref_sys',
    'pointcloud_formats',
]

SRIDS = {
    '2263': 102718
}
